from pathlib import Path
from typing import Any

import tetra3  # type: ignore
from pydantic import BaseModel, Field

from pixelemon.logging import pixelemon_LOG

TETRA_DATABASE_PATH = Path(__file__).parent / "tyc_db_to_40_deg.npz"


class TetraSolver(tetra3.Tetra3):
    def __init__(self, database_path: Path | None = None):
        if database_path is None:
            super().__init__()
        else:
            super().__init__(load_database=database_path.as_posix())
        self.settings = TetraSettings.model_validate(self.database_properties)

    @classmethod
    def low_memory(cls) -> "TetraSolver":
        solver = cls()
        pixelemon_LOG.info("Configured TetraSolver for low memory usage")
        return solver

    @classmethod
    def high_memory(cls) -> "TetraSolver":
        solver = cls(TETRA_DATABASE_PATH)
        pixelemon_LOG.info("Configured TetraSolver for high memory usage")
        return solver


class TetraSettings(BaseModel):
    dimmest_star_magnitude: float = Field(
        ...,
        description="The dimmest star magnitude to consider in the plate solve",
        alias="star_max_magnitude",
    )
    max_field_of_view: float = Field(
        ...,
        description="The maximum field of view in degrees the solver can reliably handle",
        alias="max_fov",
    )
    min_field_of_view: float = Field(
        ...,
        description="The minimum field of view in degrees the solver can reliably handle",
        alias="min_fov",
    )
    verification_star_count: int = Field(
        ...,
        description="The number of stars to use for verification of the plate solve",
        alias="verification_stars_per_fov",
    )

    def model_post_init(self, _: Any) -> None:
        pixelemon_LOG.info(f"Tetra dimmest star magnitude set to {self.dimmest_star_magnitude}")
        pixelemon_LOG.info(f"Tetra max field of view set to {self.max_field_of_view} degrees")
        pixelemon_LOG.info(f"Tetra min field of view set to {self.min_field_of_view} degrees")
        pixelemon_LOG.info(f"Tetra verification star count set to {self.verification_star_count}")


class PlateSolve(BaseModel):
    right_ascension: float = Field(
        ...,
        description="Right Ascension in degrees",
        alias="RA",
    )
    declination: float = Field(
        ...,
        description="Declination in degrees",
        alias="Dec",
    )
    roll: float = Field(
        ...,
        description="Roll angle in degrees",
        alias="Roll",
    )
    estimated_horizontal_fov: float = Field(
        ...,
        description="Estimated horizontal field of view in degrees",
        alias="FOV",
    )
    root_mean_square_error: float = Field(
        ...,
        description="Root mean square error of the plate solve in arcseconds",
        alias="RMSE",
    )
    number_of_stars: int = Field(
        ...,
        description="Number of stars used in the plate solve",
        alias="Matches",
    )
    false_positive_probability: float = Field(
        ...,
        description="Probability of a false positive plate solve",
        alias="Prob",
    )
    solve_time: float = Field(
        ...,
        description="Time taken to perform the plate solve in milliseconds",
        alias="T_solve",
    )


TETRA_SOLVER = TetraSolver.low_memory()



from math import sqrt

from pydantic import BaseModel, Field

from pixelemon.constants import MICRO_TO_MILLI


class BaseSensor(BaseModel):

    x_pixel_count: int = Field(..., description="The number of pixels in the horizontal direction")
    y_pixel_count: int = Field(..., description="The number of pixels in the vertical direction")
    pixel_width: float = Field(..., description="The width of each pixel in micrometers")
    pixel_height: float = Field(..., description="The height of each pixel in micrometers")

    @property
    def sensor_width(self) -> float:
        return self.x_pixel_count * self.pixel_width * MICRO_TO_MILLI

    @property
    def sensor_height(self) -> float:
        return self.y_pixel_count * self.pixel_height * MICRO_TO_MILLI

    @property
    def sensor_diagonal(self) -> float:
        return sqrt((self.sensor_width * self.sensor_width) + (self.sensor_height * self.sensor_height))

    @property
    def pixel_diagonal(self) -> float:
        return sqrt((self.pixel_width * self.pixel_width) + (self.pixel_height * self.pixel_height))



from pydantic import Field

from pixelemon.sensors._base_sensor import BaseSensor


class IMX296(BaseSensor):
    x_pixel_count: int = Field(default=1456, frozen=True)
    y_pixel_count: int = Field(default=1088, frozen=True)
    pixel_width: float = Field(default=3.4, frozen=True)
    pixel_height: float = Field(default=3.4, frozen=True)



import numpy as np
from numpy import typing as npt
from pydantic import BaseModel, Field, RootModel


class Detection(BaseModel):
    x_centroid: float = Field(..., description="X coordinate of the detection centroid in pixels")
    y_centroid: float = Field(..., description="Y coordinate of the detection centroid in pixels")
    semi_major_axis: float = Field(..., description="Semi-major axis of the detection in pixels")
    semi_minor_axis: float = Field(..., description="Semi-minor axis of the detection in pixels")
    angle_to_horizon: float = Field(..., description="Orientation angle of the detection in degrees")
    total_flux: float = Field(..., description="Total flux of the detection in arbitrary units")

    @classmethod
    def from_sep_object(cls, obj: dict) -> "Detection":
        return cls(
            x_centroid=obj["x"],
            y_centroid=obj["y"],
            semi_major_axis=obj["a"],
            semi_minor_axis=obj["b"],
            angle_to_horizon=obj["theta"],
            total_flux=obj["flux"],
        )


class Detections(RootModel[list[Detection]]):
    @classmethod
    def from_sep_extract(cls, objects: list[dict]) -> "Detections":
        detections = [Detection.from_sep_object(obj) for obj in objects]
        return cls(detections)

    def __len__(self) -> int:
        return len(self.root)

    def __getitem__(self, index: int) -> Detection:
        return self.root[index]

    def __iter__(self):
        return iter(self.root)

    def append(self, detection: Detection) -> None:
        self.root.append(detection)

    @property
    def y_x_array(self) -> npt.NDArray[np.float32]:
        return np.array([[det.y_centroid, det.x_centroid] for det in self.root], dtype=np.float32)



from pixelemon.optics._base_optical_assembly import BaseOpticalAssembly
from pixelemon.optics._arducam_m2508zh02 import ArducamM2508ZH02
from pixelemon.optics._arducam_cs2325zm01 import ArducamCS2325ZM01
from pixelemon.optics._arducam_cs2316zm02 import ArducamCS2316ZM02

__all__ = ["BaseOpticalAssembly", "ArducamM2508ZH02", "ArducamCS2325ZM01", "ArducamCS2316ZM02"]



from pydantic import BaseModel, Field

DEFAULT_BACKGROUND_MESH_COUNT = 40
MIN_BACKGROUND_MESH_COUNT = 10
MAX_BACKGROUND_MESH_COUNT = 100

DEFAULT_BACKGROUND_FILTER_SIZE = 2
MIN_BACKGROUND_FILTER_SIZE = 1
MAX_BACKGROUND_FILTER_SIZE = 10

DEFAULT_BACKGROUND_DETECTION_THRESHOLD = 1.0
MIN_BACKGROUND_DETECTION_THRESHOLD = 0.1
MAX_BACKGROUND_DETECTION_THRESHOLD = 6.0


class BackgroundSettings(BaseModel):
    mesh_count: int = Field(
        default=DEFAULT_BACKGROUND_MESH_COUNT,
        description="Number of meshes for background estimation",
        ge=MIN_BACKGROUND_MESH_COUNT,
        le=MAX_BACKGROUND_MESH_COUNT,
    )
    filter_size: int = Field(
        default=DEFAULT_BACKGROUND_FILTER_SIZE,
        description="Size of the filter for background smoothing",
        ge=MIN_BACKGROUND_FILTER_SIZE,
        le=MAX_BACKGROUND_FILTER_SIZE,
    )
    detection_threshold: float = Field(
        default=DEFAULT_BACKGROUND_DETECTION_THRESHOLD,
        description="Threshold for background detection",
        ge=MIN_BACKGROUND_DETECTION_THRESHOLD,
        le=MAX_BACKGROUND_DETECTION_THRESHOLD,
    )



from math import atan, degrees

from pydantic import BaseModel, Field


class BaseOpticalAssembly(BaseModel):

    image_circle_diameter: float = Field(..., description="The diameter of the image circle in millimeters")
    focal_length: float = Field(..., description="The focal length of the optical assembly in millimeters")
    focal_ratio: float = Field(..., description="The focal ratio (f-number) of the optical assembly")

    @property
    def field_of_view(self) -> float:
        return degrees(2 * atan(self.image_circle_diameter / (2 * self.focal_length)))



from pydantic import Field

from pixelemon.optics._base_optical_assembly import BaseOpticalAssembly


class ArducamM2508ZH02(BaseOpticalAssembly):
    image_circle_diameter: float = Field(default=1 / 2.5 * 25.4, frozen=True)
    focal_length: float = Field(default=8.0, frozen=True)
    focal_ratio: float = Field(default=2.0, frozen=True)



from math import ceil, pi, sqrt

import numpy as np
import numpy.typing as npt
from pydantic import BaseModel, Field

from pixelemon.constants import FWHM_TO_SIGMA
from pixelemon.logging import pixelemon_LOG

MIN_DEBLEND_MESH_COUNT = 1
MAX_DEBLEND_MESH_COUNT = 64

DEFAULT_DETECTION_THRESHOLD_SIGMA = 3.0
POINT_SOURCE_MIN_PIXEL_COUNT = 5
POINT_SOURCE_DEBLEND_MESH_COUNT = 32
POINT_SOURCE_DEBLEND_CONTRAST = 0.0005
POINT_SOURCE_FWHM = ceil(sqrt(POINT_SOURCE_MIN_PIXEL_COUNT / pi) * 2)
POINT_SOURCE_ARRAY_SIZE = int(6 * POINT_SOURCE_FWHM * FWHM_TO_SIGMA) | 1  # Ensure odd size
STREAK_SOURCE_MIN_PIXEL_COUNT = 7
STREAK_SOURCE_DEBLEND_MESH_COUNT = 1
STREAK_SOURCE_DEBLEND_CONTRAST = 1.0


class DetectionSettings(BaseModel):

    detection_threshold_sigma: float = Field(..., description="The sigma threshold for object detection")
    min_pixel_count: int = Field(
        ...,
        description="The minimum number of connected pixels to consider an object detected",
    )
    deblend_mesh_count: int = Field(
        ...,
        description="The number of meshes used for deblending overlapping objects",
        ge=MIN_DEBLEND_MESH_COUNT,
        le=MAX_DEBLEND_MESH_COUNT,
    )
    deblend_contrast: float = Field(
        ...,
        description="The contrast ratio used for deblending overlapping objects",
        ge=0.0,
        le=1.0,
    )
    merge_small_detections: bool = Field(
        ...,
        description="Whether to merge small detections with nearby larger ones",
    )

    @classmethod
    def point_source_defaults(cls) -> "DetectionSettings":
        pixelemon_LOG.info(f"Point source detection threshold set to {DEFAULT_DETECTION_THRESHOLD_SIGMA} sigma")
        pixelemon_LOG.info(f"Point source minimum pixel count set to {POINT_SOURCE_MIN_PIXEL_COUNT} pixels")
        pixelemon_LOG.info(f"Point source deblend mesh count set to {POINT_SOURCE_DEBLEND_MESH_COUNT} meshes")
        pixelemon_LOG.info(f"Point source deblend contrast set to {POINT_SOURCE_DEBLEND_CONTRAST}")
        pixelemon_LOG.info(f"Point source array size set to {POINT_SOURCE_ARRAY_SIZE} pixels")
        pixelemon_LOG.info(f"Point source FWHM set to {POINT_SOURCE_FWHM}")
        return cls(
            detection_threshold_sigma=DEFAULT_DETECTION_THRESHOLD_SIGMA,
            min_pixel_count=POINT_SOURCE_MIN_PIXEL_COUNT,
            deblend_mesh_count=POINT_SOURCE_DEBLEND_MESH_COUNT,
            deblend_contrast=POINT_SOURCE_DEBLEND_CONTRAST,
            merge_small_detections=False,
        )

    @property
    def gaussian_kernel(self) -> npt.NDArray[np.float32]:
        ax = np.array(POINT_SOURCE_ARRAY_SIZE) - POINT_SOURCE_ARRAY_SIZE // 2
        xx, yy = np.meshgrid(ax, ax)
        sigma = POINT_SOURCE_FWHM * FWHM_TO_SIGMA
        kernel = np.exp(-(xx**2 + yy**2) / (2.0 * sigma**2))
        return kernel / kernel.sum()



from pathlib import Path

from pixelemon._plate_solve import TETRA_SOLVER
from pixelemon._telescope import Telescope
from pixelemon._telescope_image import TelescopeImage

PIXELEMON_DIRECTORY = Path(__file__).parent.resolve()

__all__ = ["Telescope", "TelescopeImage", "TETRA_SOLVER"]



from pixelemon.processing._background_settings import BackgroundSettings, MIN_BACKGROUND_MESH_COUNT
from pixelemon.processing._detection_settings import DetectionSettings
from pixelemon.processing._detection import Detection, Detections

__all__ = ["BackgroundSettings", "MIN_BACKGROUND_MESH_COUNT", "DetectionSettings", "Detection", "Detections"]



from pydantic import Field

from pixelemon.constants import INCHES_TO_MILLIMETERS
from pixelemon.optics._base_optical_assembly import BaseOpticalAssembly


class ArducamCS2316ZM02(BaseOpticalAssembly):
    image_circle_diameter: float = Field(default=1 / 2.3 * INCHES_TO_MILLIMETERS, frozen=True)
    focal_length: float = Field(default=16.0, frozen=True)
    focal_ratio: float = Field(default=1.2, frozen=True)



import logging

pixelemon_LOG = logging.getLogger(__name__)
pixelemon_LOG.setLevel(logging.INFO)


# set logger name
pixelemon_LOG.name = "pixelemon"

# format time to yyyy-mm-ddThh:mm:ss.sss
formatter = logging.Formatter("%(asctime)s - %(levelname)s - %(message)s")
handler = logging.StreamHandler()
handler.setFormatter(formatter)
pixelemon_LOG.addHandler(handler)



from pathlib import Path

import pytest

from pixelemon import Telescope, TelescopeImage
from pixelemon.optics import ArducamCS2325ZM01, ArducamM2508ZH02
from pixelemon.sensors import IMX296


@pytest.fixture
def raw_fits_path() -> Path:
    return Path("tests/test-imx296-8mm.fits")


@pytest.fixture
def telescope() -> Telescope:
    sensor = IMX296()
    optical_assembly = ArducamM2508ZH02()
    telescope = Telescope(sensor=sensor, optics=optical_assembly)
    return telescope


@pytest.fixture
def image(telescope: Telescope, raw_fits_path: Path) -> TelescopeImage:
    return TelescopeImage.from_fits_file(raw_fits_path, telescope)


def test_imx296_arducam_8mm_m12(telescope: Telescope):
    assert telescope.sensor == IMX296()
    assert telescope.optics == ArducamM2508ZH02()
    assert telescope.horizontal_field_of_view == pytest.approx(31.589, abs=0.001)
    assert telescope.vertical_field_of_view == pytest.approx(23.605, abs=0.001)
    assert telescope.diagonal_field_of_view == pytest.approx(39.434, abs=0.001)
    assert telescope.horizontal_pixel_scale == pytest.approx(78.104, abs=0.001)


def test_imx296_arducam_25mm_cs():
    sensor = IMX296()
    optical_assembly = ArducamCS2325ZM01()
    telescope = Telescope(sensor=sensor, optics=optical_assembly)

    assert telescope.sensor == sensor
    assert telescope.optics == optical_assembly
    assert telescope.horizontal_field_of_view == pytest.approx(11.166, abs=0.001)
    assert telescope.vertical_field_of_view == pytest.approx(8.344, abs=0.001)
    assert telescope.diagonal_field_of_view == pytest.approx(13.939, abs=0.001)
    assert telescope.horizontal_pixel_scale == pytest.approx(27.609, abs=0.001)


def test_detections(image: TelescopeImage):
    image.crop(10.0)
    image.remove_background()
    detections = image.detections
    assert len(detections) > 0


def test_load_fits_file(telescope: Telescope, raw_fits_path: Path):
    image = TelescopeImage.from_fits_file(raw_fits_path, telescope)
    assert image._original_array is not None
    assert image._processed_array is not None
    assert image.telescope is not None
    assert image._original_array.shape[0] == image.telescope.sensor.y_pixel_count
    assert image._original_array.shape[1] == image.telescope.sensor.x_pixel_count


def test_write_fits_file(image: TelescopeImage, tmp_path: Path):
    output_path = tmp_path / "output.fits"
    image.write_to_fits_file(output_path)
    assert output_path.exists()



from pydantic import BaseModel, Field

from pixelemon.constants import DEGREES_TO_SECONDS, MICRO_TO_MILLI
from pixelemon.optics import BaseOpticalAssembly
from pixelemon.sensors import BaseSensor


class Telescope(BaseModel):
    sensor: BaseSensor = Field(..., description="The sensor used in the imaging system")
    optics: BaseOpticalAssembly = Field(..., description="The optical assembly of the imaging system")

    @property
    def diagonal_field_of_view(self) -> float:
        full_fov = self.optics.field_of_view
        ratio = self.sensor.sensor_diagonal / self.optics.image_circle_diameter
        return full_fov * ratio

    @property
    def horizontal_field_of_view(self) -> float:
        full_fov = self.optics.field_of_view
        ratio = self.sensor.sensor_width / self.optics.image_circle_diameter
        return full_fov * ratio

    @property
    def vertical_field_of_view(self) -> float:
        full_fov = self.optics.field_of_view
        ratio = self.sensor.sensor_height / self.optics.image_circle_diameter
        return full_fov * ratio

    @property
    def horizontal_pixel_scale(self) -> float:
        fov = self.horizontal_field_of_view
        ratio = self.sensor.pixel_width * MICRO_TO_MILLI / self.sensor.sensor_width
        deg_per_pixel = fov * ratio
        return deg_per_pixel * DEGREES_TO_SECONDS

    @property
    def pixel_scale(self) -> float:
        fov = self.diagonal_field_of_view
        ratio = self.sensor.pixel_diagonal * MICRO_TO_MILLI / self.sensor.sensor_diagonal
        deg_per_pixel = fov * ratio
        return deg_per_pixel * DEGREES_TO_SECONDS

    @property
    def aspect_ratio(self) -> float:
        return self.sensor.sensor_width / self.sensor.sensor_height



from pathlib import Path

import numpy as np
import numpy.typing as npt
import plotly.express as px  # type: ignore
import sep  # type: ignore
from astropy.io import fits  # type: ignore
from astropy.visualization import AsinhStretch, ZScaleInterval  # type: ignore
from pydantic import BaseModel, Field

from pixelemon._plate_solve import TETRA_SOLVER, PlateSolve
from pixelemon._telescope import Telescope
from pixelemon.constants import PERCENT_TO_DECIMAL
from pixelemon.logging import pixelemon_LOG
from pixelemon.processing import MIN_BACKGROUND_MESH_COUNT, BackgroundSettings, Detections, DetectionSettings

# from PIL import Image  # type: ignore


class TelescopeImage(BaseModel):
    _original_array: npt.NDArray[np.float32] | None = None
    _processed_array: npt.NDArray[np.float32] | None = None
    telescope: Telescope | None = None
    image_scale: float = Field(default=1.0, description="The image scale due to cropping")
    background_settings: BackgroundSettings = Field(default_factory=BackgroundSettings)
    detection_settings: DetectionSettings = Field(default_factory=DetectionSettings.point_source_defaults)

    @classmethod
    def from_fits_file(cls, file_path: Path, telescope: Telescope) -> "TelescopeImage":
        with fits.open(file_path) as hdul:
            img = cls()
            img.telescope = telescope
            img._original_array = getattr(hdul[0], "data").astype(np.float32)
            assert img._original_array is not None
            actual_ratio = img._original_array.shape[1] / img._original_array.shape[0]
            if not np.isclose(img.telescope.aspect_ratio, actual_ratio, rtol=1e-2):
                pixelemon_LOG.warning("Trimming image to match expected aspect")
                new_width = int(img._original_array.shape[0] * img.telescope.aspect_ratio)
                start_x = (img._original_array.shape[1] - new_width) // 2
                img._original_array = img._original_array[:, start_x : start_x + new_width]
                img._original_array = np.ascontiguousarray(img._original_array)
            assert img._original_array is not None
            img._processed_array = img._original_array.copy()
        pixelemon_LOG.info(f"Loaded {img._original_array.shape} image from {file_path}")
        return img

    def write_to_fits_file(self, file_path: Path):
        if self._processed_array is None:
            raise ValueError("Image array is not loaded.")
        fits.writeto(file_path, self._processed_array.astype("uint8"), overwrite=True)
        pixelemon_LOG.info(f"Saved processed image to {file_path}")

    def crop(self, crop_percent: float):
        if self._original_array is None:
            raise ValueError("Image array is not loaded.")

        crop_fraction = crop_percent * PERCENT_TO_DECIMAL
        height, width = self._original_array.shape
        crop_height = int(height * crop_fraction / 2)
        crop_width = int(width * crop_fraction / 2)
        self._processed_array = np.ascontiguousarray(
            self._original_array[crop_height : height - crop_height, crop_width : width - crop_width]
        )
        pixelemon_LOG.info(f"Image cropped to {self._processed_array.shape}")
        self.image_scale = self.image_scale * (1.0 - crop_fraction)
        new_fov = f"{self.horizontal_field_of_view:.2f} x {self.vertical_field_of_view:.2f} degrees"  # noqa: E231
        pixelemon_LOG.info(f"New field of view is {new_fov}")

    def get_brightest_detections(self, count: int) -> Detections:
        detections = self.detections
        sorted_detections = sorted(detections, key=lambda det: det.total_flux, reverse=True)
        return Detections(sorted_detections[:count])

    @property
    def background(self) -> sep.Background:
        if self._processed_array is None:
            raise ValueError("Image array is not loaded.")

        bw = max(MIN_BACKGROUND_MESH_COUNT, int(self._processed_array.shape[1] / self.background_settings.mesh_count))
        bh = max(MIN_BACKGROUND_MESH_COUNT, int(self._processed_array.shape[0] / self.background_settings.mesh_count))
        pixelemon_LOG.info(f"Background mesh size: {bw}x{bh}")

        return sep.Background(
            self._processed_array,
            bw=bw,
            bh=bh,
            fw=self.background_settings.filter_size,
            fh=self.background_settings.filter_size,
            fthresh=self.background_settings.detection_threshold,
        )

    def remove_background(self):
        if self._processed_array is None:
            raise ValueError("Image array is not loaded.")
        self._processed_array = self._processed_array - self.background

    def reset(self):
        if self._original_array is None:
            raise ValueError("Image array is not loaded.")
        self._processed_array = self._original_array.copy()
        self.image_scale = 1.0
        pixelemon_LOG.info("Image reset to original")

    @property
    def horizontal_field_of_view(self) -> float:
        if self.telescope is None:
            raise ValueError("Telescope is not set.")
        return self.telescope.horizontal_field_of_view * self.image_scale

    @property
    def vertical_field_of_view(self) -> float:
        if self.telescope is None:
            raise ValueError("Telescope is not set.")
        return self.telescope.vertical_field_of_view * self.image_scale

    @property
    def diagonal_field_of_view(self) -> float:
        if self.telescope is None:
            raise ValueError("Telescope is not set.")
        return self.telescope.diagonal_field_of_view * self.image_scale

    @property
    def detections(self):
        if self._processed_array is None:
            raise ValueError("Image array is not loaded.")

        objects = sep.extract(
            self._processed_array,
            thresh=self.detection_settings.detection_threshold_sigma * self.background.globalrms,
            minarea=self.detection_settings.min_pixel_count,
            filter_kernel=self.detection_settings.gaussian_kernel,
            deblend_nthresh=self.detection_settings.deblend_mesh_count,
            deblend_cont=self.detection_settings.deblend_contrast,
            clean=False,
            segmentation_map=False,
        )
        pixelemon_LOG.info(f"Detected {len(objects)} objects")

        return Detections.from_sep_extract(objects)

    def plot(self):

        if self._processed_array is None:
            raise ValueError("Image array is not loaded.")

        # Compute display limits using zscale
        interval = ZScaleInterval()
        vmin, vmax = interval.get_limits(self._processed_array)

        # Apply asinh stretch and clip
        stretch = AsinhStretch()
        data_stretched = stretch((np.clip(self._processed_array, vmin, vmax) - vmin) / (vmax - vmin))

        data = np.flipud(data_stretched)

        detections = self.detections
        h = data_stretched.shape[0]

        # Display with Plotly
        fig = px.imshow(data, color_continuous_scale="Viridis", origin="lower", title="FITS Image")
        for o in detections:

            r = o.semi_major_axis
            xc, yc = o.x_centroid, h - 1 - o.y_centroid
            fig.add_shape(
                type="circle",
                x0=xc - r,
                x1=xc + r,
                y0=yc - r,
                y1=yc + r,
                xref="x",
                yref="y",
                line=dict(color="red", width=1),
                opacity=0.6,
            )
        fig.update_layout(xaxis_title="X (pixels)", yaxis_title="Y (pixels)")
        fig.show()

    @property
    def plate_solve(self) -> PlateSolve:
        if self._processed_array is None:
            raise ValueError("Image array is not loaded.")
        if self.telescope is None:
            raise ValueError("Telescope is not set.")

        fov = f"{self.horizontal_field_of_view:.2f} x {self.vertical_field_of_view:.2f} degrees"  # noqa: E231
        pixelemon_LOG.info(f"Solving {len(self.detections)} detections and FOV of {fov}")

        return PlateSolve.model_validate(
            TETRA_SOLVER.solve_from_centroids(
                self.get_brightest_detections(TETRA_SOLVER.settings.verification_star_count).y_x_array,
                size=self._processed_array.shape,
                fov_estimate=self.diagonal_field_of_view,
                fov_max_error=self.diagonal_field_of_view * 0.1,
            )
        )
